=begin pod

=head1 NAME

Function::Validation - Functions to help you validate the signature of functions

Think of it as a quasi-type system for functional programming.

=head1 SYNOPSIS

=begin code :lang<raku>

use Function::Validation;

#validate a specific function
my sub func(Int $x, Int $y, Bool :$translate, :$rotate) {};
validate-function(&func, (Int, Int, Bool, Any), Mu); # => True
#                        ^--inputs              ^--return

# generate a reusable validator
my $validator = generate-func-validator((Int, Int, Bool, Any), Mu);
$validator(&func); # => True

my $lambda = sub (){"noparams"};
validate-function($lambda, [], Str);  # => True
validate-function($lambda, [], Mu);   # => True
validate-function($lambda, [], Any);  # => False (Any isn't a catch-all like Mu)

my $lamda_validator = generate-func-validator([], Mu);
$lambda_validator($lambda);           # => True

my $function_with_where = sub (Int $x where $x < 10) {$x}
validate-function($function_with_where, [Int], Mu); # => True

# extract the argument types from a function
my @types = function-arg-types(&func); #=> [(Int) (Int) (Bool) (Any)]

=end code

=head2 DESCRIPTION

Function::Validation validates a function you've been passed
to ensure it takes the inputs you expect and provides the output you expect.

Useful in functional programming when you don't have any Classes
and thus can't use a C<role> or class type to enforce valid input types.


=head2 BUT...

How does

=begin code
validate-function(&func, (Int, Int, Bool, Any), Mu)
=end code
differ from
=begin code
&func.signature ~~ :(Int,Int,Bool :$translate,Any :$rotate --> Mu)
=end code
?

=item The former is more self evident to readers.
=item The former doesn't require developers to have knowledge of the C<.signature> function
which new devs are unlikely to have.

How does
=begin code
function-arg-types(&func)
=end code
differ from
=begin code
&func.signature.params».type
=end code
?

=item The former makes what's happening more self-evident to readers.
=item The former doesn't require developers to have knowledge of the C<.signature> and C<.params> functions, or why you can't call C<.params> on the function directly.


How does is distinguish between
=begin code
sub foo(Str :$thing)
=end code
and
=begin code
sub bar(Str :$other)
=end code
? They can't be called the same way.

B<It doesn't!> If you know how to teach it to do this, please L<make a Pull Request|https://github.com/masukomi/Function-Validation/pulls>, with unit tests.

=head3 AUTHOR

L<masukomi|https://masukomi.org>

=head1 COPYRIGHT AND LICENSE

Copyright 2022

This library is free software; you can redistribute it and/or
modify it under the MIT license

=end pod


unit module Function::Validation:ver<1.0.1>:auth<masukomi (masukomi@masukomi.org)>;

#| a function to validate the signature of other functions
our sub validate-function(Sub $function, #= a reference to the function you want to test
						  Positional $arg_types, #= A list of the types. If unspecified use Mu
						  Mu $return_type = Mu #= The expected return type of the function.
						 ) is export returns Bool {
	 return False unless $function;
	 my $sig = $function.signature;
	 return False unless $sig.params.elems == $arg_types.elems;
	 return False unless $sig.returns ~~ $return_type;
	 return False unless function-arg-types($function) == $arg_types;
	 True
}

#| A function to generate a validator to validate the signature of functions
our sub generate-func-validator(Positional $arg_types, #= A list of the types. If unspecified use Mu
								Mu $return_type = Mu #= The expected return type of the function.
							   ) is export returns Sub {
	return sub (Sub $function) returns Bool { validate-function($function, $arg_types, $return_type ) };
}

#| a function to extract the types of the arguments of a function
our sub function-arg-types(Sub $function #= the function you want to know the argument types of
						  ) is export returns Seq {
	$function.signature.params.map({$_.type})
}

